import json
from redis import StrictRedis
import itertools
import pandas as pd
import requests

from .utils import SearchableDataFrame, EXPIRE

BASE_URL = "http://api.worldbank.org"


def _wb_url(x, key):
    return "{b}/{k}{c}format=json&per_page=1000&page={x}".format(
        b=BASE_URL, c="&" if "?" in key else "?", k=key,
        x=x
    )


def wb_get(key, expire=EXPIRE, force=False, raw=False):
    """
    Similar to `who_get`, only the things you might get here are more
    complicated. Handles nicely outputs coming from the higher level functions
    below (`wb_inds` and `wb_dataset`), even when they're empty or invalid.

    No guarantees for anything else.

    It might be worth creating ad hoc data types to optimise search.
    """

    r = StrictRedis()

    rkey = "WB|{}".format(key)

    if rkey in r:
        if force:
            del r[rkey]
        else:
            return json.loads(r[key].decode("utf-8"))

    l = []

    dc = False
    n = None

    for i in itertools.count(1):
        req = requests.get(_wb_url(i, key))

        j = req.json()

        if type(j) is list and len(j) == 2:
            if "page" not in j[0]:
                raise ValueError(
                    "Field 'page' not found. (got {})".format(j[0]))
            n = n or int(j[0]["pages"])
            v = j[1]
        elif type(j) is dict and "datacatalog" in j:
            if "page" not in j:
                raise ValueError("Field 'page' not found. (got {})".format(j))
            n = n or int(j["pages"])
            v = j["datacatalog"]
            dc = True
        else:
            raise ValueError(
                "Not sure how to handle the response: {}".format(j))

        v = v or []

        for d in v:
            if not dc and not raw:
                e = {}
                for k in d:
                    if type(d[k]) is dict and "id" in d[k] and "value" in d[k]:
                        e["{}_id".format(k)] = d[k]["id"]
                        e[k] = d[k]["value"]
                    elif type(d[k]) is list:
                        if d[k] and "id" in d[k][0] and "value" in d[k][0]:
                            e[k] = ", ".join(
                                [v["value"].strip() for v in d[k]])
                        else:
                            e[k] = ""
                    else:
                        e[k] = d[k]
            else:
                e = d
            l.append(e)

        if i == n or n == 0:
            break

    r[rkey] = json.dumps(l)
    r.expire(rkey, expire)

    return l


def wb_inds(k=None, expire=EXPIRE, force=False, raw=False):
    """This function takes ages. Needs some thought."""
    ds = wb_get("indicators", force=force, expire=expire)
    return ds if raw else SearchableDataFrame(ds)(k)


# todo: add parameters like mrv etc.
def wb_dataset(ind, countries="all", d1=1960, d2=2016, frequency="Y",
               expire=EXPIRE, force=False, raw=False):
    if countries != "all":
        countries = ";".join(countries)

    params = "date={d1}:{d2}&frequency={f}".format(d1=d1, d2=d2, f=frequency)
    k = "countries/{c}/indicators/{ind}?{params}".format(c=countries,
                                                         ind=ind,
                                                         params=params)
    ds = wb_get(k, force=force, expire=expire)
    return ds if raw else pd.DataFrame(ds)



import pandas as pd

from .who import who_dataset, who_dims, who_get
from .oecd import oecd_dataset, oecd_dims, oecd_inds, oecd_md
from .wb import wb_dataset, wb_inds, wb_get
from .eurostat import eus_dataset, eus_inds

import kleptes.goodies
from ._version import __version__

# you really need this.
pd.set_option('max_colwidth', 160)

__all__ = [
    "who_dataset", "who_dims", "who_get",
    "oecd_dataset", "oecd_dims", "oecd_inds",
    "wb_dataset", "wb_inds", "wb_get",
    "eus_dataset", "eus_inds",
    "pd",
    "__version__"
]



import uuid

import pandas as pd


def _sbs(df, ascending=True, copy=True):
    """
    Sometimes you want a stacked bar plot and have the bars sorted by their
    sum: this is the case for labelled bar plots, which are stacked plots where
    only one bar has non-zero height, e.g. say you have a metric by country,
    each of which is labelled by a region/continent and you want to colour the
    bars by region/continent. Simply call df.sbs() on your multicolumn
    dataframe and you're done. If you only have one column this function will
    have no effect.

    :param df: a DataFrame (1d index only and, ideally, many columns)
    :param ascending: sort in ascending order (default: False)
    :param copy: get a copy or sort in-place (default: True)
    :return: your sorted DataFrame
    """
    if type(df) is not pd.DataFrame:
        raise TypeError("The argument must be a DataFrame")
    _df = df.copy() if copy else df
    c = ""
    while True:
        c = uuid.uuid4()
        if c not in _df.columns:
            break
    _df[c] = _df.sum(axis=1)
    _df = _df.sort_values(c, ascending=ascending)
    del _df[c]
    return _df


pd.DataFrame.sbs = _sbs



import json
import re

import pandas as pd
import requests
from redis import StrictRedis

from kleptes.utils import EXPIRE, get_re


def reformat_numbers(d):
    for f in {"a_land_area_kmsq_2012", "d_year"}:
        if f in d:
            n = d[f].replace(",", "")
            d[f] = int(n) if "." not in n else float(n)


def lddf(l, full=False):
    if full:
        d = []
        for r in l:
            if "attr" in r:
                a = {"a_" + d["category"].lower(): d["value"] for d in
                     r["attr"]}
                reformat_numbers(a)
                r.pop("attr")
                d.append(dict(r, **a))
        return pd.DataFrame(d)
    else:
        return pd.DataFrame(l)[["label", "display"]]


def pmatch(l, pattern):
    rxp = get_re(pattern)
    return [{"label": k["label"],
             "display": k["display"]} for k in l if
            re.match(rxp, k["label"]) or re.match(rxp, k[
                "display"])]


def parse_ds(ds, names=True):
    l = []

    if names:
        countries = who_countries()
        regions = who_regions()
    else:
        countries, regions = {}, {}

    for x in ds:
        e = {}
        dim = {"d_" + d["category"].lower(): d["code"] for d in x["Dim"]}
        val = {"v_{k}".format(k=k): x["value"][k] for k in
               {"high", "low", "numeric"} if k in x["value"]}
        reformat_numbers(dim)
        x.pop("Dim")
        x.pop("value")
        e.update(x)
        e.update(dim)
        e.update(val)

        if names:
            if "d_country" in e:
                e["country"] = countries.get(e["d_country"])
            if "d_region" in e:
                e["region"] = regions.get(e["d_region"])

        l.append(e)
    return pd.DataFrame(l)


def who_get(string="", filter_=None, force=False, expire=EXPIRE,
            raw=False):
    """
    Low-level function to get JSON datasets from the WHO database.

    :param string: the thingy to be appended to the url (e.g. 'GHO/blah').
    :param filter_: a filter as per API support (ugh underscore)
    :param force: ignores cached items (but re-caches them)
    :param expire: self-explanatory
    :param raw: returns the JSON dataset as a string (useful to debug stuff).
    :return: the requested data set
    """
    filter_ = "&filter={}".format(filter_) if filter_ is not None else ""
    url = "http://apps.who.int/gho/athena/api/{}?format=json{}".format(string,
                                                                       filter_)
    r = StrictRedis()
    key = "WHO|{}|{}".format(string, filter_)

    if not force and r.exists(key):
        t = r.get(key).decode("utf-8")
        if raw:
            return t
        else:
            return json.loads(t)
    else:
        req = requests.get(url)

        # to be removed once they fix the API call result
        t = req.text.replace("Radio band\"\"", "Radio band\"")

        r[key] = t
        if expire is not None:
            r.expire(key, expire)
        if raw:
            return t
        j = json.loads(t)
        return j


def who_dims(name=None, force=False, expire=EXPIRE, full=False):
    """
    Get the dimensions from the WHO database. You can do things like
    `who_dims("*gh*")`.

    :param name: the dimension's name (as mentioned, you can search with `*`)
    :param force: ignores cached items (but re-caches them)
    :param expire: self-explanatory
    :param full: if False it only returns label and full name).
    :return: the requested data set
    """
    dims = who_get()["dimension"]
    if name is None or name == "":
        return lddf(dims, full)
    elif "*" not in name:
        return lddf(dims, full).query("label == @name or display == @name")
    else:
        return lddf(
            pmatch(who_get("", force=force, expire=expire)["dimension"], name),
            full)


def who_dataset(dim=None, name=None, filter_=None, force=False,
                expire=EXPIRE, raw=False, parse=True, full=False):
    """
    Get the dataset "dim/name" from the WHO database. The cool thing is that
    you can do pattern search on both dimension and name, although if you do it
    on the dimension the name parameter is ignored.

    If you fix a dimension (ideally "GHO" since that's where most of the stuff
    is) you can do stuff like `who_dataset("GHO", "*alcohol*")` and it will
    search for all the datasets with matching descriptions so no
    need to actually remember the ID's.

    :param dim: the dimension's name
    :param name: the dataset's name
    :param filter_: a filter as per API support
    :param force: ignores cached items (but re-caches them)
    :param expire: self-explanatory
    :param raw: returns the JSON dataset as a raw string (useful to debug stuff).
    :param parse: returns a nice DataFrame as a result (probably what you want unless you're debugging)
    :param full: returns the full dataset (by default it only has label and full name).
    :return: the requested data set
    """
    dims = who_dims(name=dim, force=force, expire=expire)
    if len(dims) > 1:
        return dims
    elif len(dims) == 1:
        dim = dims.iloc[0]["label"]
        if name is None or name == "":
            ds = who_get("{dim}/".format(dim=dim), force=force, expire=expire)[
                "dimension"][0]["code"]
        elif "*" in name:
            ds = pmatch(
                who_get("{dim}/".format(dim=dim), force=force, expire=expire)[
                    "dimension"][0]["code"], name)
        else:
            # exact match
            ds = who_get("{dim}/{name}/".format(dim=dim, name=name),
                         filter_=filter_, force=force, expire=expire, raw=raw)
            return parse_ds(ds["fact"]) if parse and not raw else ds
        return lddf(ds, full=full) if not raw else ds


# two handy shortcuts
def who_countries():
    return {d["label"]: d["display"] for d in who_dataset("COUNTRY", raw=True)}


def who_regions():
    return {d["label"]: d["display"] for d in who_dataset("REGION", raw=True)}



from io import BytesIO
import gzip

import pandas as pd
from redis import StrictRedis
import requests
import numpy as np

from .utils import SearchableDataFrame, EXPIRE


def _eus_proc_tsv(s, key, expire=EXPIRE, force=False):
    """
    This thing is horrific. It's meant to completely reprocess the data
    not just so that it's 'better' for pandas, but also so that it's
    actually readable.

    Datasets are shaped this way:

    ``"pk1,pk2,...,pkn\time"\t"yyyyMmm"\t"yyyyMmm"...``

    So that:

    1. The first field is always a comma-separated primary key (lol)
    2. Years are not part of the primary key, they're columns (lol)

    This function outputs a DataFrame shaped as:

    ``pk1;pk2;...;pkn;time;value``

    with the primary key split accordingly and time as a row index.
    """
    key = "EUS_ds|{}|proc".format(key)
    r = StrictRedis()
    if force:
        if key in r:
            del r[key]

    if key in r:
        return pd.read_csv(BytesIO(r[key]), compression="gzip")

    df = None
    ndf = None
    cols = None

    hl = 0

    l = []

    for row in gzip.open(BytesIO(s)):
        row = row.decode("utf-8")
        ra = row.strip().split("\t")

        # get the "primary key"
        fe = ra[0].split(",")
        del ra[0]

        ra = fe + ra  # the row is the full primary key + the actual row
        if df is None:
            hl = len(fe)

            # of course the fourth element of the primary key is called
            # ``"pkn\time"``.
            ra[hl-1] = ra[hl-1].replace("\\time", "")
            cols = ra

            for i, k in enumerate(ra[hl:]):
                # years contain months as M. replacing to convert YYYYMmm into
                # YYYY-mm
                ra[hl + i] = k.strip().replace("M", "-")

            df = pd.DataFrame([], columns=cols)
            for k in df.columns[hl:]:
                df[k] = df[k].astype("float")
        else:
            for i, k in enumerate(ra):
                if i >= hl:
                    # of course ":" means NA
                    if ":" in k:
                        ra[i] = np.nan
                    else:
                        # sometimes floats have notes. (e.g. "195.2 f").
                        spi = k.strip().find(" ")
                        if spi > 0:
                            ra[i] = float(k.strip()[:spi])
                        else:
                            ra[i] = float(k)
            l.append(ra)
        # do it in bulks
        if len(l) >= 500:
            # the actual reshaping + type conversion
            _df = (
                df.append(pd.DataFrame(l, columns=cols))
                .set_index(cols[:hl]).stack().astype("float")
            )
            if ndf is None:
                ndf = _df
            else:
                ndf = ndf.append(_df)
            l = []

    # parse the rest. kind of c-esque, must find a better way
    if len(l) > 0:
        _df = (
            df.append(pd.DataFrame(l, columns=cols))
            .set_index(cols[:hl]).stack().astype("float")
        )
        ndf = ndf.append(_df) if ndf is not None else _df

    ndf.index.names = list(ndf.index.names[:hl]) + ["time"]
    ndf = ndf.to_frame("value").reset_index()

    r[key] = gzip.compress(ndf.to_csv(index=None).encode("utf-8"))
    r.expire(key, expire)

    return ndf


def _eus_get(key, url, expire=EXPIRE, force=False):
    r = StrictRedis()

    if force:
        del r[key]
    if key in r:
        return r[key]
    else:
        t = requests.get(url).content
        r[key] = t
        r.expire(key, expire)
        return r[key]


def eus_inds(k=None, expire=EXPIRE, force=False):
    """
    Same syntax as all the others. Returns a ``SearchableDataFrame``.

    :param k: search key (only in the title column)
    :param expire: same as usual
    :param force: same as usual
    :return: a ``SearchableDataFrame`` with the indicators and other stuff.
    """
    key = "EUS_inds"
    url = ("http://ec.europa.eu/eurostat/estat-navtree-portlet-prod/"
           "BulkDownloadListing?file=table_of_contents_en.txt")
    b = _eus_get(key, url, expire=expire, force=force)

    eudf = SearchableDataFrame(pd.read_csv(BytesIO(b), sep="\t"))
    return eudf(k, cols="title")


def eus_dataset(ds=None, expire=EXPIRE, force=False):
    """
    This thing returns a barely recognisable version of the original dataset;
    much more efficient for pandas manipulation.

    It downloads stuff from the bulk download service instead of the API
    because the API isn't searchable in any published way.

    :param ds: exact match this time
    :param expire: same as usual
    :param force: same as usual
    :return: a ``SearchableDataFrame`` with the indicators and other stuff.
    """
    cds = set(eus_inds()["code"])
    if ds not in cds:
        raise KeyError("Code '{}' not found".format(ds))

    key = "EUS_ds|{}".format(ds)
    url = ("http://ec.europa.eu/eurostat/estat-navtree-portlet-prod/"
           "BulkDownloadListing?sort=1&"
           "file=data%2F{code}.tsv.gz").format(code=ds)

    df = _eus_proc_tsv(_eus_get(key, url, expire=expire, force=force), key=key,
                       expire=expire, force=force)
    return df



from setuptools import setup
from kleptes._version import __version__


DESC = """
You know that time you wanted to get some world data but it was a dreadful
jungle of bad documentation, requests that make very little sense, ridiculously
complicated data formats (SDMX?), and in the end you just said fuck this
and decided to do something else?

This is an attempt at making world data mining suck a bit less.

Currently available data sets:

* World Bank
* OECD
* WHO
* Eurostat

Source and (minimal) documentation `here <https://github.com/myyc/kleptes/>`_.
"""


setup(
    name="kleptes",
    version=__version__,
    author="myyc",
    description="Goodies to mine world data",
    license="BSD",
    keywords="data mining python pandas who oecd",
    packages=["kleptes"],
    long_description=DESC,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
    install_requires=["requests", "pandas", "redis", "bs4"],
)



import re
import json

import requests
from bs4 import BeautifulSoup
from redis import StrictRedis
import pandas as pd

from .utils import SearchableDataFrame, get_re, EXPIRE

BASE_URL = "http://stats.oecd.org/sdmx-json"


def get_index(ds):
    """Converts the index to a DatetimeIndex"""
    v = [pd.Period(k["id"]).start_time for k in
         ds["structure"]["dimensions"]["observation"][0]["values"]]
    idx = pd.DatetimeIndex(v)
    idx.name = "d"
    return idx


def pk(ds, ks):
    """Converts things like 0:2:0:1 into a dictionary matching
    what the numbers mean."""
    a = [int(i) for i in ks.split(":")]
    key = ds["structure"]["dimensions"]["series"]
    n = [k["name"].lower() for k in key]

    # this should be configurable. right now it's just 'id': 'value',
    # could be 'name': 'value', etc.
    l = [[v["id"] for v in k["values"]] for k in key]

    return {n[i]: l[i][a[i]] for i in range(len(a))}


def parseds(ds):
    if len(ds["dataSets"]) > 1:
        print(len(ds["dataSets"]))
    s = ds["dataSets"][0]["series"]

    df = None

    for k in s:
        kt = pk(ds, k)

        o = s[k]["observations"]
        l = {int(i): o[i][0] for i in o}
        idx = get_index(ds)
        _df = (
            pd.Series(l).to_frame("value").reindex(range(0, len(idx)))
            .assign(dt=idx).assign(**kt).set_index("dt")
        )
        df = _df if df is None else pd.concat((df, _df))

    return df


class OECDims:
    def __iter__(self):
        return self._keys.__iter__()

    def __getitem__(self, i):
        return self._d[i]

    def __repr__(self):
        return "OECDims(\"{}\", {})".format(self._name, self._keys)

    def __call__(self, pattern=None):
        df = None
        for k in self:
            _df = self._d[k](pattern).assign(dim=k)
            df = _df if df is None else pd.concat([df, _df])
        return df[["dim", "id", "name"]]

    def __init__(self, md, name=""):
        self._keys = []
        self._d = {}
        self._name = "\"\"" if len(name) == 0 else name
        for x in md:
            if "role" in x and x["role"] == "TIME_PERIOD":
                continue
            attr = x["name"].lower()
            self._keys.append(attr)
            self._d[attr] = SearchableDataFrame(x["values"],
                                                columns=["id", "name"])
            self.__setattr__(attr, self._d[attr])


def oecd_inds(pattern=None, force=False, expire=EXPIRE, df=True):
    """No way to get the indicators without scraping."""
    key = "OECD|inds"
    r = StrictRedis()
    if key in r and not force:
        l = json.loads(r[key].decode("utf-8"))
    else:
        t = requests.get(BASE_URL + "/").text
        soup = BeautifulSoup(t, "html.parser")
        l = {i["value"].strip(): i.text.strip() for i in
             list(soup.find(id="Datasets").children)[2:] if i != "\n"}
        r[key] = json.dumps(l)
        r.expire(key, expire)
    if pattern is not None and pattern != "":
        rxp = get_re(pattern)
        l = {k: l[k] for k in l if
             re.match(rxp, k) or re.match(rxp, l[k])}
    if df:
        d = [{"id": k, "desc": l[k]} for k in l]
        return pd.DataFrame(d, columns=["id", "desc"])
    else:
        return l


def oecd_md(pattern, force=False, expire=EXPIRE):
    idxs = oecd_inds(pattern, df=False)
    if len(idxs) == 0:
        raise ValueError("Empty selection")
    elif len(idxs) > 1:
        raise ValueError("Selection is too wide")

    idx = list(idxs.keys())[0]
    key = "OECD|{}|md".format(idx)

    r = StrictRedis()

    if key not in r or force:
        j = requests.get("{}/metadata/{}".format(BASE_URL, idx)).json()
        r[key] = json.dumps(j)
        r.expire(key, expire)
    else:
        j = json.loads(r[key].decode("utf-8"))

    return j


def oecd_dims(pattern, force=False, expire=EXPIRE):
    md = oecd_md(pattern, force, expire)
    name = md["structure"]["name"]
    return OECDims(md["structure"]["dimensions"]["observation"], name)


def _pmise(x):
    return "+".join(x["id"])


def oecd_dataset(idx, rawstr=None, force=False, expire=EXPIRE, raw=False,
                 **kwargs):
    if rawstr is None:
        if kwargs is None:
            raise ValueError(
                "If `rawstr` is None you need to set kwargs according "
                "to the dimension.")
        dims = oecd_dims(idx)

        rawstr = ".".join([_pmise(dims[l](kwargs[l])) for l in dims])

    key = "OECD|{}|data".format(rawstr)

    r = StrictRedis()
    if key not in r or force:
        j = requests.get(
            "{}/data/{}/{}/all".format(BASE_URL, idx, rawstr)).json()
        r[key] = json.dumps(j)
        r.expire(key, expire)
    else:
        j = json.loads(r[key].decode("utf-8"))
    if raw:
        return j
    else:
        return parseds(j).assign(ind=idx.upper())



import re
import fnmatch

import pandas as pd

# three days
EXPIRE = 259200


def get_re(pattern):
    if type(pattern) == str:
        return re.compile(fnmatch.translate(pattern), flags=re.IGNORECASE)
    elif type(pattern) == list:
        return re.compile("|".join(fnmatch.translate(pattern)),
                          flags=re.IGNORECASE)
    else:
        raise ValueError("'pattern' must either be a string or a list "
                         "(or None)")


class SearchableDataFrame(pd.DataFrame):
    """A DataFrame that implements a search method via __call__"""

    def __call__(self, pattern=None, cols=None):
        if pattern is None:
            return self

        p = get_re(pattern)

        if cols not in list(self.columns):
            if cols is None:
                _df = self.select_dtypes(["object"])
            elif type(cols) == list:
                _df = self[cols]
            else:
                raise ValueError("'cols' must be either a 'list', 'None' or "
                                 "the name of a column.")
            s = _df.apply(lambda x: x.str.match(p)).sum(axis=1).astype("bool")
        else:
            s = self[cols].str.match(p)
        return self[s]
